import { hexByteStringToUint8Array, layoutDiscriminator, serializeLayout, deserializeLayout, addFixedValues, } from "@wormhole-foundation/sdk-base";
import { chainItem, universalAddressItem, signatureItem, sequenceItem, guardianSetItem, } from "./layout-items";
import { keccak256 } from "./utils";
export const composeLiteral = (protocol, payloadName) => (protocol ? `${protocol}:${payloadName}` : payloadName);
const decomposeLiteral = (payloadLiteral) => {
    const index = payloadLiteral.indexOf(":");
    return (index !== -1
        ? [payloadLiteral.slice(0, index), payloadLiteral.slice(index + 1)]
        : [null, payloadLiteral]);
};
const guardianSignatureLayout = [
    { name: "guardianIndex", binary: "uint", size: 1 },
    { name: "signature", ...signatureItem },
];
const headerLayout = [
    { name: "version", binary: "uint", size: 1, custom: 1, omit: true },
    { name: "guardianSet", ...guardianSetItem },
    {
        name: "signatures",
        binary: "array",
        lengthSize: 1,
        layout: guardianSignatureLayout,
    },
];
//envelope + payload are getting hashed and signed
const envelopeLayout = [
    { name: "timestamp", binary: "uint", size: 4 },
    { name: "nonce", binary: "uint", size: 4 },
    { name: "emitterChain", ...chainItem() },
    { name: "emitterAddress", ...universalAddressItem },
    { name: "sequence", ...sequenceItem },
    { name: "consistencyLevel", binary: "uint", size: 1 },
];
const baseLayout = [...headerLayout, ...envelopeLayout];
const payloadFactory = new Map();
function getPayloadLayout(layoutLiteral) {
    const layout = payloadFactory.get(layoutLiteral);
    if (!layout)
        throw new Error(`No layout registered for payload type ${layoutLiteral}`);
    return layout;
}
const payloadLiteralToPayloadItem = (payloadLiteral) => (payloadLiteral === "Uint8Array"
    ? { name: "payload", binary: "bytes" }
    : {
        name: "payload",
        binary: "object",
        layout: getPayloadLayout(payloadLiteral),
    });
export const createVAA = (payloadLiteral, vaaData) => {
    const [protocolName, payloadName] = decomposeLiteral(payloadLiteral);
    const bodyLayout = [
        ...envelopeLayout,
        payloadLiteralToPayloadItem(payloadLiteral),
    ];
    const bodyWithFixed = addFixedValues(bodyLayout, 
    //not sure why the unknown cast here is required and why the type isn't inferred correctly
    vaaData);
    return {
        protocolName,
        payloadName,
        payloadLiteral,
        ...addFixedValues(headerLayout, vaaData),
        ...bodyWithFixed,
        hash: keccak256(serializeLayout(bodyLayout, bodyWithFixed)),
    };
};
export function registerPayloadType(protocol, name, layout) {
    const payloadLiteral = composeLiteral(protocol, name);
    if (payloadFactory.has(payloadLiteral))
        throw new Error(`Payload type ${payloadLiteral} already registered`);
    payloadFactory.set(payloadLiteral, layout);
}
export function registerPayloadTypes(protocol, payloads) {
    for (const [name, layout] of payloads)
        registerPayloadType(protocol, name, layout);
}
export const serialize = (vaa) => {
    const layout = [
        ...baseLayout,
        payloadLiteralToPayloadItem(vaa.payloadLiteral),
    ];
    return serializeLayout(layout, vaa);
};
export const serializePayload = (payloadLiteral, payload) => {
    if (payloadLiteral === "Uint8Array")
        return payload;
    const layout = getPayloadLayout(payloadLiteral);
    return serializeLayout(layout, payload);
};
export function payloadDiscriminator(payloadLiterals, allowAmbiguous) {
    const literals = (() => {
        if (Array.isArray(payloadLiterals[0]))
            return payloadLiterals.flatMap(([protocol, payloadNames]) => payloadNames.map((name) => composeLiteral(protocol, name)));
        if (typeof payloadLiterals[1] === "string")
            return payloadLiterals;
        const [protocol, payloadNames] = payloadLiterals;
        return payloadNames.map((name) => composeLiteral(protocol, name));
    })();
    const discriminator = layoutDiscriminator(literals.map(literal => getPayloadLayout(literal)), !!allowAmbiguous);
    return ((data) => {
        if (typeof data === "string")
            data = hexByteStringToUint8Array(data);
        const cands = discriminator(data);
        return (Array.isArray(cands))
            ? cands.map((c) => literals[c])
            : cands !== null ? literals[cands] : null;
    });
}
;
export function deserialize(payloadDet, data) {
    if (typeof data === "string")
        data = hexByteStringToUint8Array(data);
    const [header, envelopeOffset] = deserializeLayout(headerLayout, data, 0, false);
    //ensure that guardian signature indicies are unique and in ascending order - see:
    //https://github.com/wormhole-foundation/wormhole/blob/8e0cf4c31f39b5ba06b0f6cdb6e690d3adf3d6a3/ethereum/contracts/Messages.sol#L121
    for (let i = 1; i < header.signatures.length; ++i)
        if (header.signatures[i].guardianIndex <=
            header.signatures[i - 1].guardianIndex)
            throw new Error("Guardian signatures must be in ascending order of guardian set index");
    const [envelope, payloadOffset] = deserializeLayout(envelopeLayout, data, envelopeOffset, false);
    const [payloadLiteral, payload] = typeof payloadDet === "string"
        ? [
            payloadDet,
            deserializePayload(payloadDet, data, payloadOffset)
        ]
        : deserializePayload(payloadDet, data, payloadOffset);
    const [protocolName, payloadName] = decomposeLiteral(payloadLiteral);
    const hash = keccak256(data.slice(envelopeOffset));
    return {
        protocolName,
        payloadName,
        payloadLiteral,
        ...header,
        ...envelope,
        payload,
        hash
    };
}
export function deserializePayload(payloadDet, data, offset = 0) {
    //grouped together to have just a single cast on the return type
    return (() => {
        if (typeof data === "string")
            data = hexByteStringToUint8Array(data);
        if (payloadDet === "Uint8Array")
            return data.slice(offset);
        if (typeof payloadDet === "string")
            return deserializeLayout(getPayloadLayout(payloadDet), data, offset);
        //kinda unfortunate that we have to slice here, future improvement would be passing an optional
        //  offset to the discriminator
        const candidate = payloadDet(data.slice(offset));
        if (candidate === null)
            throw new Error(`Encoded data does not match any of the given payload types - ${data}`);
        return [
            candidate,
            deserializeLayout(getPayloadLayout(candidate), data, offset),
        ];
    })();
}
export const blindDeserializePayload = (() => {
    const rebuildDiscrimininator = () => {
        const layoutLiterals = Array.from(payloadFactory.keys());
        const layouts = layoutLiterals.map((l) => payloadFactory.get(l));
        return [layoutLiterals, layoutDiscriminator(layouts, true)];
    };
    let layoutLiterals = [];
    let discriminator = (_) => [];
    return (data) => {
        if (payloadFactory.size !== layoutLiterals.length)
            [layoutLiterals, discriminator] = rebuildDiscrimininator();
        if (typeof data === "string")
            data = hexByteStringToUint8Array(data);
        const candidates = discriminator(data).map((c) => layoutLiterals[c]);
        return candidates.reduce((acc, literal) => {
            try {
                acc.push([literal, deserializePayload(literal, data)]);
            }
            catch { }
            return acc;
        }, []);
    };
})();
//# sourceMappingURL=vaa.js.map